class DecodeError(ValueError):
    def __init__(self, string, position, message):
        line = string.count("\n", 0, position) + 1
        column = position - string.rfind("\n", 0, position)
        super_message = (
            f"{message} at line {line} column {column} (position {position})"
        )
        super().__init__(super_message)
        self.string = string
        self.message = message
        self.position = position
        self.line = line
        self.column = column


class ExpectedBracketError(DecodeError):
    def __init__(self, string, position, bracket, message=None):
        if message is None:
            message = f"Expected bracket `{bracket}`"
        super().__init__(string, position, message)
        self.bracket = bracket


class ExpectedDelimiterError(DecodeError):
    def __init__(self, string, position, delimiter, message=None):
        if message is None:
            message = f"Expected delimiter `{delimiter}`"
        super().__init__(string, position, message)
        self.delimiter = delimiter


class ExpectedKeyError(DecodeError):
    def __init__(self, string, position, message="Expected key"):
        super().__init__(string, position, message)


class ExpectedValueError(DecodeError):
    def __init__(self, string, position, message="Expected value"):
        super().__init__(string, position, message)


class ExtraneousDataError(DecodeError):
    def __init__(
        self,
        string,
        position,
        message="Extraneous data (likely indicating a malformed JSON document)",
    ):
        super().__init__(string, position, message)


class UnexpectedCharError(DecodeError):
    def __init__(self, string, position, char, message=None):
        if message is None:
            message = f"Unexpected char `{char}`"
        super().__init__(string, position, message)
        self.char = char


class UnknownCharError(DecodeError):
    def __init__(self, string, position, char, message=None):
        if message is None:
            message = f"Unknown char `{char}`"
        super().__init__(string, position, message)
        self.char = char


class UnmatchedBracketError(DecodeError):
    def __init__(self, string, bracket, position, message=None):
        if message is None:
            message = f"Unmatched bracket `{bracket}`"
        super().__init__(string, position, message)
        self.bracket = bracket


class ScanStringError(DecodeError):
    def __init__(self, string, position, message=None):
        if message is None:
            message = "Error scanning string"
        super().__init__(string, position, message)


class INIDecodeError(DecodeError):
    def __init__(self, string, position, message):
        super_message = f"{message} at position {position}"
        super(ValueError, self).__init__(super_message)
        self.string = string
        self.message = message
        self.position = position



import json
import json.scanner
import math
import re
from datetime import datetime
from functools import partial
from typing import Any, Callable, Dict, List, Tuple, Union

import arrow
import dateutil

from .exc import (
    ExpectedBracketError,
    ExpectedDelimiterError,
    ExpectedKeyError,
    ExpectedValueError,
    ExtraneousDataError,
    ScanStringError,
    UnexpectedCharError,
    UnknownCharError,
    UnmatchedBracketError,
)
from .obj import JSONObject


__all__ = ["Scanner"]


Number = Union[int, float]


WHITESPACE = " \f\n\r\t\v"
WHITESPACE_RE = re.compile(r"[ \f\n\r\t\v]*")


def skip_whitespace(
    string,
    i,
    *,
    comments=True,
    whitespace=WHITESPACE,
    whitespace_re=WHITESPACE_RE,
) -> int:
    if string[i : i + 1] in whitespace:
        i = whitespace_re.match(string, i).end()
    if comments and string[i : i + 2] == "//":
        i = string.find("\n", i + 3)
        i = len(string) if i == -1 else i + 1
        next_char = string[i : i + 1]
        if next_char in whitespace or (comments and next_char == "/"):
            return skip_whitespace(string, i, comments=comments)
    return i


def scan_object(
    string,
    i,
    *,
    scan,
    stack,
    stack_push,
    stack_pop,
    enable_extras=True,
    converter=JSONObject,
    skip_chars=WHITESPACE,
    skip_whitespace=skip_whitespace,
) -> Tuple[Union[Dict, Any], int]:
    if string[i : i + 1] != "{":
        raise ExpectedBracketError(string, i, "{")

    i += 1

    if string[i : i + 1] in skip_chars:
        i = skip_whitespace(string, i, comments=enable_extras)

    if string[i : i + 1] == "}":
        return converter() if converter else {}, i + 1

    obj = {}
    stack_push(("{", i - 1))

    while True:
        # Key
        if string[i : i + 1] != '"':
            raise ExpectedKeyError(string, i)

        key, i = scan(string, i)

        # Delimiting colon, which may be followed by whitespace
        if string[i : i + 1] == ":":
            i += 1
            if string[i : i + 1] in skip_chars:
                i = skip_whitespace(string, i, comments=enable_extras)
        else:
            raise ExpectedDelimiterError(string, i, ":")

        # Value
        value, i = scan(string, i)

        # Add entry
        obj[key] = value

        # Comma, which may be followed by whitespace
        if string[i : i + 1] == ",":
            comma_i = i
            i += 1
            if string[i : i + 1] in skip_chars:
                i = skip_whitespace(string, i, comments=enable_extras)
        else:
            comma_i = None

        if not string[i : i + 1]:
            break

        # Closing brace
        if string[i : i + 1] == "}":
            if comma_i is not None and not enable_extras:
                raise UnexpectedCharError(string, comma_i, ",")
            stack_pop(string, "{", "}", i)
            i += 1
            break

    return converter(**obj) if converter else obj, i


def scan_array(
    string,
    i,
    *,
    scan,
    stack,
    stack_push,
    stack_pop,
    enable_extras=True,
    skip_chars=WHITESPACE,
    skip_whitespace=skip_whitespace,
) -> Tuple[List, int]:
    if string[i : i + 1] != "[":
        raise ExpectedBracketError(string, i, "[")

    i += 1

    if string[i : i + 1] in skip_chars:
        i = skip_whitespace(string, i, comments=enable_extras)

    if string[i : i + 1] == "]":
        return [], i + 1

    array = []
    array_append = array.append
    stack_push(("[", i - 1))

    while True:
        # Value
        value, i = scan(string, i)
        array_append(value)

        # Comma, which may be followed by whitespace
        if string[i : i + 1] == ",":
            comma_i = i
            i += 1
            if string[i : i + 1] in skip_chars:
                i = skip_whitespace(string, i, comments=enable_extras)
        else:
            comma_i = None

        if not string[i : i + 1]:
            break

        if string[i : i + 1] == "]":
            if comma_i is not None and not enable_extras:
                raise UnexpectedCharError(string, comma_i, ",")
            stack_pop(string, "[", "]", i)
            i += 1
            break

    return array, i


scan_string = json.decoder.scanstring


# NOTE: To avoid ambiguity with ints, dates must have at least a year
#       and month part and times must have at least an hour and minute
#       part.
YEAR = r"\d\d\d[1-9]"
MONTH = r"0[1-9]|1[1-2]"
DAY = r"[0-2][1-9]|3[0-1]"
HOUR = r"[0-2]\d"
MINUTE = r"[0-5]\d"
SECOND = r"[0-5]\d"
MICROSECOND = r"\d{1,6}"
DATE = rf"{YEAR}-({MONTH})(-({DAY}))?"
TIME_ZONE = rf"Z|[+-]{HOUR}:{MINUTE}"
TIME = rf"{HOUR}:{MINUTE}(:{SECOND}(\.{MICROSECOND})?)?"
DATE_TIME = rf"{DATE}T{TIME}(?P<tz>{TIME_ZONE})?"
# Regex, time only flag
# NOTE: The order of these items matters!
DATETIME_CONVERTERS = (
    (re.compile(DATE_TIME), False),
    (re.compile(DATE), False),
    (re.compile(TIME), True),
)

TZ_LOCAL = dateutil.tz.tzlocal()


def scan_date(
    string,
    i,
    *,
    today,
    converters=DATETIME_CONVERTERS,
    # NOTE: Arrow requires this to be a *list*
    time_formats=["HH:mm", "HH:mm:ss", "HH:mm:ss.S"],
    tz_local=TZ_LOCAL,
) -> [datetime, int]:
    for (regex, is_time_only) in converters:
        match = regex.match(string, i)
        if match is not None:
            end = match.end()
            str_val = string[i:end]
            if is_time_only:
                try:
                    val = arrow.get(str_val, time_formats)
                except ValueError:
                    raise ValueError(f"Could not convert matched time: {str_val}")
                val = val.replace(
                    year=today.year,
                    month=today.month,
                    day=today.day,
                    tzinfo=today.tzinfo,
                )
            else:
                args = {}
                if "tz" not in match.groupdict():
                    args["tzinfo"] = tz_local
                try:
                    val = arrow.get(str_val, **args)
                except ValueError:
                    raise ValueError(f"Could not convert matched date/time: {str_val}")
            val = val.datetime
            return val, end
    return None


DECIMAL = r"[0-9](_?[0-9]+)*"
FLOAT_WITH_EXP = rf"{DECIMAL}(\.{DECIMAL})?[eE][+-]?{DECIMAL}"
FLOAT_WITHOUT_EXP = rf"{DECIMAL}\.{DECIMAL}"
FLOAT = rf"[+-]?({FLOAT_WITH_EXP}|{FLOAT_WITHOUT_EXP})"

# Regex, converter (const or callable), const flag
# NOTE: The const flag is used to avoid the function call overhead of
#       checking `callable(converter)`
# NOTE: The order of these items matters!
NUMBER_CONVERTERS = (
    # Constants
    (re.compile(r"([+-])?(inf|Infinity)"), math.inf, True),
    (re.compile(r"([+-])?(nan|NaN)"), math.nan, True),
    (re.compile(r"([+-])?(E)"), math.e, True),
    (re.compile(r"([+-])?(π|PI)"), math.pi, True),
    (re.compile(r"([+-])?(τ|TAU)"), math.tau, True),
    # Floats
    (re.compile(FLOAT), float, False),
    # Integers
    (re.compile(r"[+-]?0[bB]_?[0-1](_?[0-1]+)*"), partial(int, base=2), False),
    (re.compile(r"[+-]?0[oO]_?[0-7](_?[0-7]+)*"), partial(int, base=8), False),
    (re.compile(r"[+-]?0[xX]_?[0-f](_?[0-9a-f]+)*"), partial(int, base=16), False),
    (re.compile(r"[+-]?((0(_?0+)*|[1-9](_?[0-9]+)*))"), int, False),
)


def scan_number(
    string,
    i,
    *,
    converters=NUMBER_CONVERTERS,
) -> [Number, int]:
    for (regex, converter, is_const) in converters:
        match = regex.match(string, i)
        if match is not None:
            end = match.end()
            str_val = string[i:end]
            if is_const:
                pre = match.groups()[0] or "+"
                val = converter if pre == "+" else -converter
            else:
                try:
                    val = converter(str_val)
                except ValueError:
                    raise ValueError(f"Could not convert matched number: {str_val}")
            return val, end
    return None


class Scanner:

    """Scan JSONish string and return a Python object."""

    def __init__(
        self,
        *,
        strict=True,
        prescan=None,
        scan_object=scan_object,
        object_converter=JSONObject,
        scan_array=scan_array,
        scan_string=scan_string,
        scan_date=scan_date,
        scan_number=scan_number,
        enable_extras=True,
        fallback_scanner=None,
    ):
        self.prescan = prescan
        self.scan_object = scan_object
        self.object_converter = object_converter
        self.scan_array = scan_array
        self.scan_string = scan_string
        self.scan_date = scan_date
        self.scan_number = scan_number
        self.fallback_scanner = fallback_scanner
        self.enable_extras = enable_extras
        self.strict = strict
        self.stack = []
        # Ensure all bare times use the same today value
        self.today = arrow.now(tz=TZ_LOCAL).floor("day")
        self.scan = self.make_scan_method()

    def decode(self, string, *, ignore_extra_data=False) -> Union[Any, Tuple[Any, int]]:
        """Scan JSONish string and return a Python object.

        When creating a :class:`Decoder` for scanning multiple JSON
        documents, this is the method that should generally be used.
        This method cleans up internal state between scans whereas
        :meth:`scan` does not.

        .. note:: :class:`Decoder` is an alias for :class:`Scanner`. The
            former is preferred for most public usage.

        """
        self.stack.clear()
        obj, i = self.scan(string, start=True)
        if ignore_extra_data:
            return obj, i
        elif i != len(string):
            raise ExtraneousDataError(string, i)
        return obj

    def make_scan_method(self) -> Callable[[str, int], Tuple[Any, int]]:
        def stack_pop(
            string,
            left,
            right,
            right_i,
            *,
            stack=self.stack,
            pop=self.stack.pop,
        ):
            if not stack:
                raise UnmatchedBracketError(string, right, right_i)
            top, left_i = pop()
            if top != left:
                raise UnmatchedBracketError(string, left, left_i)

        def scan(
            string,
            i=0,
            *,
            start=False,
            # Instance config
            strict=self.strict,
            prescan=self.prescan,
            scan_object=self.scan_object,
            object_converter=self.object_converter,
            scan_array=self.scan_array,
            scan_string=self.scan_string,
            scan_date=self.scan_date,
            scan_number=self.scan_number,
            enable_extras=self.enable_extras,
            fallback_scanner=self.fallback_scanner,
            skip_chars=(WHITESPACE + "/" if self.enable_extras else WHITESPACE),
            skip_whitespace=skip_whitespace,
            # Instance vars
            stack=self.stack,
            stack_push=self.stack.append,
            stack_pop=stack_pop,
            today=self.today,
            # Other locals
            no_val=object(),
            default_scan_number=json.scanner.NUMBER_RE.match,
        ) -> Tuple[Any, int]:
            if start and not string[i:]:
                if enable_extras:
                    return None, i
                raise ExpectedValueError(string, i, message="JSON document is empty")

            if string[i : i + 1] in skip_chars:
                i = skip_whitespace(string, i, comments=enable_extras)

            if not string[i:]:
                raise ExpectedValueError(string, i)

            val = no_val
            char = string[i]

            if prescan is not None:
                result = prescan(self, scan, string, i)
                if result is not None:
                    val, i = result
                    # XXX: This duplicates code below because of early
                    #      return.
                    if start and char in "{[" and stack:
                        bracket, position = stack[-1]
                        raise UnmatchedBracketError(string, bracket, position)
                    if string[i : i + 1] in skip_chars:
                        i = skip_whitespace(string, i, comments=enable_extras)
                    return val, i

            if char == "{":
                val, i = scan_object(
                    string,
                    i,
                    scan=scan,
                    stack=stack,
                    stack_push=stack_push,
                    stack_pop=stack_pop,
                    enable_extras=enable_extras,
                    converter=object_converter,
                    skip_chars=skip_chars,
                    skip_whitespace=skip_whitespace,
                )

            elif char == "[":
                val, i = scan_array(
                    string,
                    i,
                    scan=scan,
                    stack=stack,
                    stack_push=stack_push,
                    stack_pop=stack_pop,
                    enable_extras=enable_extras,
                    skip_chars=skip_chars,
                    skip_whitespace=skip_whitespace,
                )

            elif char == '"':
                try:
                    val, i = scan_string(string, i + 1, strict)
                except json.JSONDecodeError as exc:
                    raise ScanStringError(string, exc.pos, exc.msg)

            elif char == "n" and string[i : i + 4] == "null":
                val, i = None, i + 4

            elif char == "t" and string[i : i + 4] == "true":
                val, i = True, i + 4

            elif char == "f" and string[i : i + 5] == "false":
                val, i = False, i + 5

            elif enable_extras:
                if char in "0123456789":
                    result = scan_date(string, i, today=today)
                    if result is not None:
                        val, i = result

                if val is no_val and char in "0123456789+-iInNEPπTτ":
                    result = scan_number(string, i)
                    if result is not None:
                        val, i = result

                if val is no_val and fallback_scanner:
                    result = fallback_scanner(self, scan, string, i)
                    if result is not None:
                        val, i = result

            elif char in "0123456789-":
                match = default_scan_number(string, i)
                if match is not None:
                    integer, fraction, exponent = match.groups()
                    if fraction or exponent:
                        val = float(f"{integer}{fraction or ''}{exponent or ''}")
                    else:
                        val = int(integer)
                    i = match.end()

            if val is no_val:
                raise UnknownCharError(string, i, char)

            if start and char in "{[" and stack:
                bracket, position = stack[-1]
                raise UnmatchedBracketError(string, bracket, position)

            if string[i : i + 1] in skip_chars:
                i = skip_whitespace(string, i, comments=enable_extras)

            return val, i

        return scan



# flake8: noqa

from .exc import DecodeError, DecodeError as JSONDecodeError

# Decode from JSON to Python
from .decoder import (
    decode,
    decode as loads,
    decode_file,
    decode_file as load,
    Decoder,
    Decoder as JSONDecoder,
)

# Encode from Python to JSON
from .encoder import (
    encode,
    encode as dumps,
    encode_to_file,
    encode_to_file as dump,
    Encoder,
    Encoder as JSONEncoder,
)



"""JSONish Decoder

In addition to standard JSON, the JSONish decoder also supports/handles
the following:

- Trailing commas

- Line comments starting with //

- Any valid Python int or float:
  - Literal binary, octal, and hex values
  - Underscore separators in numbers
  - Unary plus operator

- Math constants:
  - inf, nan, E, π, PI, τ, TAU
  - Infinity, NaN

- Empty input strings will be converted to ``None`` rather than raising
  an exception.

- Literal (unquoted) dates & times:
  - 2021-06
  - 2021-06-23
  - 2021-06-23T12:00
  - 2021-06-23T12:00Z
  - 2021-06-23T12:00-07:00
  - 12:00 (today's date at noon)

.. note:: For dates and times, when a time zone isn't specified, the
    local time zone will be used.

- An object converter can be specified to convert JSON objects into a
  custom type (or types). By default, JSON objects will be converted to
  :class:`scanner.JSONObject`s, which allows properties to be accessed
  with either dotted or bracket notation.

- Loading config from INI files with values encoded as JSON.

- *All* scanning methods can be overridden if some additional
  customization is required.

- A prescan method can be provided to handle values before the JSON
  scanners are applied.

- A fallback scanner method can be provided to handle additional types
  of values if none of the default scanners are suitable.

.. note:: See details below in :func:`decode` for using prescan or a
    fallback scanner.

Examples::

    >>> decode("+1")
    1
    >>> decode("1_000")
    1000
    >>> decode("[0b11]")
    [3]
    >>> d = decode("2021-06-24")
    >>> d.timetuple()[:5]
    (2021, 6, 24, 0, 0)
    >>> d.tzinfo
    tzlocal()
    >>> decode(" [1, 2 ,  3  ,   ] ")
    [1, 2, 3]
    >>> decode("[[]]")
    [[]]

"""
from configparser import ConfigParser, ExtendedInterpolation
from pathlib import Path
from typing import Any, Callable, Optional, TextIO, Tuple, Union

from . import scanner
from .exc import INIDecodeError
from .obj import JSONObject
from .scanner import Scanner, Scanner as Decoder


__all__ = ["decode", "decode_file", "Decoder"]


# Signature for prescanner and fallback scanner functions
AltScannerFunc = Callable[[Scanner, Callable, str, int], Tuple[Any, int]]


def decode(
    string: str,
    *,
    strict: bool = True,
    prescan: Optional[AltScannerFunc] = None,
    scan_object: Callable = scanner.scan_object,
    object_converter: Callable = JSONObject,
    scan_array: Callable = scanner.scan_array,
    scan_string: Callable = scanner.scan_string,
    scan_date: Callable = scanner.scan_date,
    scan_number: Callable = scanner.scan_number,
    fallback_scanner: Optional[AltScannerFunc] = None,
    enable_extras: bool = True,
    ignore_extra_data: bool = False,
    ini=False,
) -> Union[Any, Tuple[Any, int]]:
    """Scan JSONish string and return a Python object.

    - JSON object -> Python object (see below)
    - JSON array -> Python list
    - JSON string -> Python string
    - Literal date string (no quotes) -> Python datetime
    - JSON number -> Python number (int or float)
    - Empty string -> None

    By default, JSON objects are converted to simple Python namespace
    objects that allow attributes to be accessed via dotted or bracket
    notation. These objects can be converted to plain dicts with
    ``dict(obj)`` or you can use ``object_converter=None`` to get back
    plain dicts.

    A different ``object_converter`` can be passed to customize object
    creation, perhaps based on a type field::

        def converter(obj):
            if "__type__" in obj:
                # Convert to type based on __type__
                T = types[obj["__type__"]]
                return T(**obj)
            # Don't convert since no type was specified
            return obj

    When errors are encountered, various kinds of exceptions are
    thrown. These all derive from :class:`DecodeError`, which in turn
    derives from the builtin :class:`ValueError`.

    Examples::

        >>> import arrow, math

        >>> decode("") is None
        True

        >>> d = decode("2021-06-23")
        >>> d.timetuple()[:5]
        (2021, 6, 23, 0, 0)

        >>> t = arrow.now()
        >>> d = decode("12:00")
        >>> d.timetuple()[:5] == (t.year, t.month, t.day, 12, 0)
        True

        >>> d = decode("2021-06-23T12:00")
        >>> d.timetuple()[:6]
        (2021, 6, 23, 12, 0, 0)

        >>> decode("[inf, nan]")
        [inf, nan]

        >>> decode("E") == math.e
        True
        >>> (decode("π"), decode("PI")) == (math.pi, math.pi)
        True
        >>> (decode("τ"), decode("TAU")) == (math.tau, math.tau)
        True

        >>> decode("0"), decode("+0"), decode("-0"), decode("000")
        (0, 0, 0, 0)
        >>> decode("1"), decode("+1"), decode("-1")
        (1, 1, -1)

        >>> decode("1.0"), decode("+1.0"), decode("-1.0")
        (1.0, 1.0, -1.0)

        >>> decode("0b11"), decode("0o11"), decode("0x11")
        (3, 9, 17)

        >>> decode("{}", object_converter=None), decode("[]")
        ({}, [])

        >>> decode("[0b11, 11, 0x11]")
        [3, 11, 17]

    When the ``ignore_extra_data`` flag is set, a tuple will be returned
    containing 1) a Python object representing the part of the JSON
    string that was successfully parsed and 2) the index in the JSON
    string where the extra data starts. In most cases, extra data
    indicates an error, but this flag can be used to intentionally
    include extra data:

        >>> decode('{} # ignored', object_converter=None, ignore_extra_data=True)
        ({}, 3)

    There are a couple of advanced/esoteric/low level features for use
    where additional customization of parsing is required:

    - The prescanner. This is a callable that takes the scanner
      instance, the primary scan function, the JSON input string, and
      the current position; it can either return a value and the next
      position or ``None``. Returning ``None`` indicates that the
      prescanner didn't handle the string and that the regular scanners
      should be tried.

    - The fallback scanner. This is a callable that takes a scanner
      instance, the primary scan function, the JSON input string, and
      the current position; it must return a Python value along with
      the next position.

    """
    instance = scanner.Scanner(
        strict=strict,
        prescan=prescan,
        scan_object=scan_object,
        object_converter=object_converter,
        scan_array=scan_array,
        scan_string=scan_string,
        scan_date=scan_date,
        scan_number=scan_number,
        enable_extras=enable_extras,
        fallback_scanner=fallback_scanner,
    )
    if ini:
        return decode_ini(string, instance)
    return instance.decode(string, ignore_extra_data=ignore_extra_data)


def decode_ini(string, scanner):
    """Decode INI file with JSONish values.

    INI section and setting names are split on dots to create sub-dicts.
    For example::

        [section.one]
        a.b = 1

    will result in the following dict::

        {"section": "one": {"a": {"b": 1}}}

    To disable splitting a name, wrap it in parentheses. For example::

        [(section.one)]
        (a.b) = 1

    will result in the following dict::

        {"section.one": {"a.b": 1}}

    """
    result = {}
    parser = ConfigParser(interpolation=ExtendedInterpolation())
    parser.optionxform = lambda option: option
    parser.read_string(string)

    for section_name in parser.sections():
        obj = result
        section_path = parse_ini_name(section_name)
        for segment in section_path[:-1]:
            if segment not in obj:
                obj[segment] = {}
            obj = obj[segment]
        obj[section_path[-1]] = section_config = {}

        section_items = parser[section_name]
        for raw_name, raw_value in section_items.items():
            obj = section_config
            path = parse_ini_name(raw_name)
            for segment in path[:-1]:
                if segment not in obj:
                    obj[segment] = {}
                obj = obj[segment]
            value, _ = scanner.scan(raw_value)
            obj[path[-1]] = value

    return result


def parse_ini_name(name):
    """Parse section and setting names from INI file."""
    # NOTE: Groups can't be nested, so this is simplistic

    if "(" not in name and ")" not in name:
        return name.split(".")

    j = 0
    segments = []
    last_right_paren = 0

    while True:
        i = name.find("(", j)
        if i == -1:
            break

        unmatched_i = name.find(")", j)
        if -1 < unmatched_i < i:
            raise INIDecodeError(name, unmatched_i, f"Unmatched ) in name '{name}'")

        before = name[j:i]
        if before:
            if not before.endswith("."):
                raise INIDecodeError(name, i - 1, f"Expected dot in name '{name}'")
            segments.append(before[:-1])

        j = name.find(")", i + 1)
        if j == -1:
            raise INIDecodeError(name, i, f"Unmatched ( in name '{name}'")
        last_right_paren = j

        nested_i = name.find("(", i + 1)
        if -1 < nested_i < j:
            raise INIDecodeError(name, nested_i, f"Nested ( in name '{name}'")

        part = name[i + 1 : j]
        segments.append(part)

    unmatched_i = name.find(")", last_right_paren + 1)
    if unmatched_i != -1:
        raise INIDecodeError(name, unmatched_i, f"Unmatched ) in name '{name}'")

    after = name[j + 1 :]
    if after:
        if not after.startswith("."):
            raise INIDecodeError(name, j + 1, f"Expected dot in name '{name}'")
        segments.append(after[1:])

    return segments


def decode_file(
    file: Union[str, Path, TextIO],
    *,
    strict: bool = True,
    prescan: Optional[AltScannerFunc] = None,
    scan_object: Callable = scanner.scan_object,
    object_converter: Callable = JSONObject,
    scan_array: Callable = scanner.scan_array,
    scan_string: Callable = scanner.scan_string,
    scan_date: Callable = scanner.scan_date,
    scan_number: Callable = scanner.scan_number,
    fallback_scanner: Optional[AltScannerFunc] = None,
    enable_extras: bool = True,
    ignore_extra_data: bool = False,
    ini=None,
) -> Union[Any, Tuple[Any, int]]:
    """Read file, scan JSONish string, and return a Python object.

    This reads the file into a string, then calls :func:`decode`; see
    its docstring for details.

    TODO: This reads the whole file all at once. Maybe find a way to
          stream the file contents, although this would add a bit of
          complexity to the scanner in order to make it generically
          handle iterables of chars.

    """
    if isinstance(file, str):
        path = Path(file)
        with path.open() as fp:
            string = fp.read()
    elif isinstance(file, Path):
        path = file
        with file.open() as fp:
            string = fp.read()
    else:
        path = None
        string = file.read()
    if ini is None:
        if path is not None and path.suffix in (".cfg", ".ini", ".inijson"):
            ini = True
        else:
            ini = False
    return decode(
        string,
        strict=strict,
        prescan=prescan,
        scan_object=scan_object,
        object_converter=object_converter,
        scan_array=scan_array,
        scan_string=scan_string,
        scan_date=scan_date,
        scan_number=scan_number,
        fallback_scanner=fallback_scanner,
        enable_extras=enable_extras,
        ignore_extra_data=ignore_extra_data,
        ini=ini,
    )



import types


__all__ = ["JSONObject"]


class JSONObject(types.SimpleNamespace):
    """A simple bucket of attributes.

    JSON objects will be converted to this type by default.

    .. note:: In cases where you generally need/want to work with dicts,
        .pass ``None`` as the ``object_converter`` to :func:`decode`,
        .:class:`Scanner`, etc.

    Items can be accessed via dotted or bracket notation::

        >>> obj = JSONObject(x=1)
        >>> obj.x
        1
        >>> obj["x"]
        1

    An object can be converted to a dict by calling `dict` on it::

        >>> obj = JSONObject(x=1)
        >>> dict(obj)
        {'x': 1}

    """

    def __getitem__(self, name):
        return self.__dict__.__getitem__(name)

    def __setitem__(self, name, value):
        self.__dict__[name] = value

    def __iter__(self):
        return iter(self.__dict__.items())

    def __repr__(self):
        return repr(dict(self))

    def __str__(self):
        return str(dict(self))



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['jsun']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['jsun = jsun.__main__:main']}

setup_kwargs = {
    'name': 'jsun',
    'version': '1.0a1',
    'description': 'JSON encoding & decoding with extra features',
    'long_description': '# jsun\n\nThis is an alternative JSON decoder/encoder that supports some extra\nfeatures. It takes a *lot* of inspiration from the `json` module in the\nstandard library and exposes the same high level API: `load`, `loads`,\n`dump`, `dumps`, `JSONDecoder`, and `JSONEncoder`.\n\nIn many cases, `jsun` can be swapped in by installing the package then\nsimply updating imports to use `jsun` instead of `json`.\n\n## Extra decoding features\n\n- Trailing commas\n \n- Line comments starting with //\n \n- All valid Python ints and floats:\n  - Binary, octal, hex\n  - Underscore separators\n  - Unary plus operator\n\n- Math constants:\n  - inf, nan, E, π, PI, τ, TAU\n  - Infinity, NaN\n\n- Literal (unquoted) dates and times:\n  - 2021-06\n  - 2021-06-23\n  - 2021-06-23T12:00\n  - 2021-06-23T12:00Z\n  - 2021-06-23T12:00-07:00\n  - 12:00 (today\'s date at noon)\n\n- Decoding an empty string will produce `None` rather than an exception\n  (an exception will be raised if extras are disabled)\n\n- *All* parsing methods can be overridden if some additional\n  customization is required. In particular, the object and array\n  parsers can be overridden\n\n- A pre-parse method can be provided to handle values before the regular\n  JSON parsers are applied\n\n- A fallback parsing method can be provided to handle additional types\n  of values if none of the default parsers are suitable\n\n- When errors are encountered, specific exceptions are raised (all\n  derived from the built-in `ValueError`)\n\n## Extra encoding features\n\nThe `jsun` encoder is very similar to the standard library encoder (and\nis in fact a subclass of `json.JSONEncoder`). Currently, it supports\nonly a couple of extra features:\n\n- Date objects are converted to ISO format by default\n- Datetime objects are converted to ISO format by default\n\nNOTE: There is some asymmetry here. E.g., date and datetime objects\nshould be converted to literals instead of quoted strings.\n\n## Disabling the extra features\n\n*All* the extra features can be turned off with a flag:\n\n    >>> from jsun import decode\n    >>> decode("[1, 2, 3,]")\n    [1, 2, 3]\n    >>> decode("[1, 2, 3,]", enable_extras=False)\n    <exception traceback>\n\n## Differences between jsun and standard library json\n\n- An empty string input is converted to `None` rather than raising an\n  exception (only if extras are enabled).\n\n- When decoding, instead of `object_hook` and `object_hook_pairs`,\n  there\'s just a single `object_converter` argument. It\'s essentially\n  the same as `object_hook`. `object_hook_pairs` seems unnecessary\n  nowadays since `dict`s are ordered.\n\n- The default object type is `jsun.obj.JSONObject` instead of `dict`. A\n  `JSONObject` is a bucket of properties that can be accessed via dotted\n  or bracket notation. Pass `object_converter=None` to get back\n  `dict`s instead.\n\n## Config files\n\nA bonus feature is that configuration can be loaded from INI files\nwhere the keys are split on dots to create sub-objects and the values\nare encoded as JSON.\n\nThis is quite similar to TOML and some of the features of `jsun`, like\nliteral dates, are inspired by TOML.\n\nThis feature was originally developed in 2014 as part of the\n`django-local-settings` project, about a year and half after TOML was\nfirst released but before I\'d heard of it.\n\n### Differences with TOML\n\n- Parentheses are used instead of quotes to avoid splitting on dots\n- Objects created using `{}` syntax (AKA "inline tables" in TOML) can\n  span multiple lines\n- There are no arrays of tables\n- Others I\'m not thinking of at the moment...\n\n## About the name\n\nMy first choice was `jsonish` but that\'s already taken. My second choice\nwas `jsonesque` but it\'s also taken, and it\'s hard to type. `jsun` is\nnice because it\'s easy to type and easy to swap in for `json` by just\nchanging a single letter.\n\n## Testing\n\nThere\'s a suite of unit tests, which also tests against the JSON checker\nfiles at https://json.org/JSON_checker/. Coverage is currently at 82%.\n',
    'author': 'Wyatt Baldwin',
    'author_email': 'self@wyattbaldwin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wylee/jsun',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import argparse
import pathlib
import sys

from .decoder import decode, decode_file
from .encoder import encode, encode_to_file


def main(argv=None):
    parser = argparse.ArgumentParser(prog="jsun")

    type_group = parser.add_mutually_exclusive_group(required=True)
    type_group.add_argument("-j", "--json", help="JSON string to decode")
    type_group.add_argument("-f", "--file", help="File to decode")

    parser.add_argument(
        "-c",
        "--ini",
        action="store_true",
        default=None,
        help="Force INI+JSON",
    )

    parser.add_argument("-o", "--out-file", help="Write decoded JSON to file")

    parser.add_argument(
        "-i",
        "--indent",
        type=int,
        default=None,
        help="Pretty print with specified indent level",
    )

    parser.add_argument(
        "-d",
        "--disable-extras",
        dest="enable_extras",
        action="store_false",
        default=True,
    )

    args = parser.parse_args(argv)
    json = args.json
    file = args.file
    out_file = args.out_file
    ini = args.ini
    indent = args.indent
    enable_extras = args.enable_extras

    if json:
        obj = decode(json, enable_extras=enable_extras, ini=ini)
    elif file:
        obj = decode_file(file, enable_extras=enable_extras, ini=ini)

    if out_file:
        path = pathlib.Path(out_file)
        with path.open("w") as fp:
            encode_to_file(obj, fp, indent=indent, enable_extras=enable_extras)
    else:
        print(encode(obj, indent=indent, enable_extras=enable_extras))


if __name__ == "__main__":
    sys.exit(sys.argv)



"""JSONish Encoder

This passes through to the stdlib `json` module with handling of
:class:`JSONObject` and :class:`datetime.datetime` values.

"""
import datetime
import json

from .obj import JSONObject


__all__ = ["encode", "encode_to_file", "Encoder"]


class Encoder(json.JSONEncoder):
    def __init__(
        self,
        *,
        skipkeys=False,
        ensure_ascii=True,
        check_circular=True,
        allow_nan=True,
        sort_keys=False,
        indent=None,
        separators=None,
        default=None,
        enable_extras=True,
    ):
        super().__init__(
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            sort_keys=sort_keys,
            indent=indent,
            separators=separators,
            default=default,
        )
        self.enable_extras = enable_extras

    def default(self, obj):
        if self.enable_extras:
            if isinstance(obj, JSONObject):
                return dict(obj)
            # NOTE: datetime objects are also date objects
            if isinstance(obj, datetime.datetime):
                return obj.isoformat()
            if isinstance(obj, datetime.date):
                return obj.isoformat()
        # Raise TypeError
        return super().default(obj)


def encode(
    obj,
    *,
    cls=Encoder,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    sort_keys=False,
    indent=None,
    separators=None,
    default=None,
    enable_extras=True,
):
    instance = cls(
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        sort_keys=sort_keys,
        indent=indent,
        separators=separators,
        default=default,
        enable_extras=enable_extras,
    )
    return instance.encode(obj)


def encode_to_file(
    obj,
    fp,
    *,
    cls=Encoder,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    sort_keys=False,
    indent=None,
    separators=None,
    default=None,
    enable_extras=True,
):
    instance = cls(
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        sort_keys=sort_keys,
        indent=indent,
        separators=separators,
        default=default,
        enable_extras=enable_extras,
    )
    iterable = instance.iterencode(obj)
    for chunk in iterable:
        fp.write(chunk)
